#
#    Copyright 2001,2004 Greenwood Genetic Center
#
#    This file is part of Linkage Scripts.
#
#    Linkage Scripts is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Linkage Scripts is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Linkage Scripts; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
package       Linkage;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK $version);

require       Exporter;
@ISA        = qw(Exporter);
@EXPORT     = qw(&parseDatafile &parsePedfile %datainfo @pedinfo);
@EXPORT_OK  = qw($datafile $pedfile);
$version    = 0.10;

use vars      qw($datafile $pedfile %datainfo @pedinfo);

# use the variables we plan to export
$datafile = "datain.dat";
$pedfile  = "pedin.dat";

# parse the datafile
sub parseDatafile {
  my($linetxt, $lineno, $i, $markerfreq, $totalfreq, $loci, @locisort);

  open(DATAFILE, "<$datafile")
      or die "Couldn't open $datafile: $!\n";

  # 1st line - no loci, risk locus, sexlinked (if 1), program code
  $linetxt = <DATAFILE>;
  if($linetxt =~ m/^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*(<<.*)?$/) {
    $datainfo{"nlocus"} = $1;

    if($2 > $datainfo{"nlocus"} || $2 < 0) {
      die("$datafile,1: the risk locus must be 0 or " .
                  "the disease locus if risk is to be calculated\n");
    }
    $datainfo{"risklocus"} = $2;

    if($3 != 1 && $3 != 0) {
      die("$datafile,1: sex-linked must be 0 or 2\n");
    }
    $datainfo{"sexlink"} = $3;

    if($4 < 1 || $4 > 7) {
      die("$datafile,1: the program code must be a number 1-7\n" .
                  "             1 = CILINK\n" .
                  "             2 = CMAP\n" .
                  "             3 = ILINK\n" .
                  "             4 = LINKMAP\n" .
                  "             5 = MLINK\n" .
                  "             6 = LODSCORE\n" .
                  "             7 = CLODSCORE\n");
    }
    $datainfo{"nprogram"} = $4;
  } else {
    die("$datafile,1: 4 numbers must be present\n");
  }

  # 2nd line - mut locus, mut male, mut fem, hap freq (if 1)
  $linetxt = <DATAFILE>;
  if($linetxt =~ m/^\s*(\d+)\s+([\d.]+)\s+([\d.]+)\s+(\d+)\s*(<<.*)?$/) {
    if($1 > $datainfo{"nlocus"} || $2 < 0) {
      die("$datafile,2: the mutation locus must be 0 or " .
                  "the mutation locus\n");
    }
    $datainfo{"mutsys"} = $1;

    if($2 < 0 || $2 > 1) {
      die("$datafile,2: the male mutation rate must be " .
          "between 0 and 1\n");
    }
    $datainfo{"mutmale"} = $2;

    if($3 < 0 || $3 > 1) {
      die("$datafile,2: the female mutation rate must be " .
          "between 0 and 1\n");
    }
    $datainfo{"mutfem"} = $3;

    if($4 != 0) {
      if($4 == 1) {
        die("$datafile,2: locus disequilibrium is not implemented\n");
      } else {
        die("$datafile,2: the locus disequilibrium does not equal 1 or 0\n");
      }
    }
    $datainfo{"disequil"} = $4;
  } else {
    die("$datafile,2: 4 numbers must be present\n");
  }

  # 3rd line - order of loci
  $linetxt = <DATAFILE>;
  if($linetxt =~ m/^\s*((\d+\s+)+\d+)\s*(<<.*)?$/) {
    $i = 1;
    $datainfo{"order"} = [split(/\s+/, $1)];
    for($i = 0; $i < $datainfo{"nlocus"}; $i++) {
      $locisort[$datainfo{"order"}->[$i] - 1] = $datainfo{"order"}->[$i];
    }
    for($i = 0; $i < $datainfo{"nlocus"}; $i++) {
      if((! exists($locisort[$i])) || $i + 1 != $locisort[$i]) {
        die("$datafile,3: not all loci present on line 3\n");
      }
    }
  } else {
    die("$datafile,3: lociorder not present on line 3\n");
  }

  $lineno = 3;

  for($loci = 1; $loci <= $datainfo{"nlocus"}; $loci++) {
    $lineno++;
    $linetxt = <DATAFILE>;
    if($linetxt =~ m/^\s*(\d+)\s+(\d+)\s*(<<.*)?$/) {
      if($1 != 1 && $1 != 3) {
        die("$datafile,4: 1st number must be the type of trait\n" . 
            "  0 - Quantitive variable (not implemented)\n" .
            "  1 - Affection status\n" .
            "  2 - Binary factors (not implemented)\n" .
            "  3 - Numbered alleles\n");
      }
      $datainfo{"$loci"} = [[$1, $2]];

      if($linetxt =~ m/<<.*#\d+\s+(\S+(\s\S+)*)\s*$/) {
        $datainfo{"names"}->[$loci - 1] = $1;
      }
    } else {
      die("$datafile,$lineno: 2 numbers must be listed\n");
    }

    # get the actual frequencies
    $lineno++;
    $linetxt = <DATAFILE>;
    if($linetxt =~ m/^\s*(([\d.]+\s+)+[\d.]+)\s*(<<.*)?$/) {
      $i = 0;
      $totalfreq= 0;
      foreach(split(/\s+/, $1)) {
        $i++;
        $totalfreq+= $_;
      }
      if($i != $datainfo{"$loci"}->[0]->[1]) {
        die("$datafile,$lineno: the listed number of frequencies " . 
            "($markerfreq) does not match the total number ($i) " .
            "frequencies for locus $loci\n");
      }
      if($totalfreq < 0.9999999999 || $totalfreq > 1.0000000001) {
        die("$datafile,$lineno: loci frequencies do not add up to 1\n"); 
      }
      if($linetxt =~ m/<<.*#\d+\s+(\S+(\s\S+)*)\s*$/) {
        $datainfo{"names"}->[$loci - 1] = $1;
      } else {
        if(! exists($datainfo{"names"}->[$loci])) {
          $datainfo{"names"}->[$loci - 1] = $loci;
        }
      }
    } else {
        die("$datafile,$lineno: loci frequencies for marker $loci were " .
            "not found.\n");
    }

    if($datainfo{"$loci"}->[0]->[0] == 1) {
      $lineno++;
      $linetxt = <DATAFILE>;
      if($linetxt =~ m/^\s*(\d+)\s*(<<.*)?$/) {
        if($datainfo{"sexlink"} == 1 && $1 != 1) {
          warn("$datafile,$lineno: multiple liability classes with sexlinked " .
               "data is might not be supported correctly.\n");
        }
        $datainfo{"$loci"}->[1] = [$1];
      } else {
        die("$datafile,$lineno: 1 number must be present\n");
      }

      for($i = 1; $i <= $1; $i++) {
        $lineno++;
        $linetxt = <DATAFILE>;
        if($linetxt =~ m/^\s*([\d.]+)\s+([\d.]+)\s+([\d.]+)\s*(<<.*)?$/) {
          if($datainfo{"sexlink"} == 1) {
            $datainfo{"$loci"}->[$i * 2 - 1] = [$1, $2, $3];
          } else {
            $datainfo{"$loci"}->[$i] = [$1, $2, $3];
          }
        } else {
          die("$datafile,$lineno: 3 numbers for the liability class are " .
              "required\n");
        }
        if($datainfo{"sexlink"} == 1) {
          $lineno++;
          $linetxt = <DATAFILE>;
          if($linetxt =~ m/^\s*([\d.]+)\s+([\d.]+)\s*(<<.*)?$/) {
            $datainfo{"$loci"}->[$i * 2] = [$1, $2];
          } else {
            die("$datafile,$lineno: 2 numbers for the male liability class " .
                "are required\n");
          }
        }
      }
    }
  }

  # sex difference, interference
  $lineno++;
  $linetxt = <DATAFILE>;
  if($linetxt =~ m/^\s*(\d+)\s+(\d+)\s*(<<.*)?$/) {
    if($1 != 0) {
      die("$datafile,$lineno: sex interference should be equal to 0\n");
    }
    $datainfo{"sexdiff"} = $1;

    if($2 != 0) {
      die("$datafile,$lineno: sex interference should be equal to 0\n");
    }
    $datainfo{"sexintf"} = $2;
  } else {
    die("$datafile,$lineno: 3 numbers must be present\n");
  }

  # get the loci distances
  $lineno++;
  $linetxt = <DATAFILE>;
  if($linetxt =~ m/^\s*(([\d.]+\s+)*[\d.]+)\s*(<<.*)?$/) {
    $datainfo{"recomb"} = [split(/\s+/, $1)];
    if(@{$datainfo{"recomb"}} != $datainfo{"nlocus"} - 1) {
      die("$datafile," . (9 + ($datainfo{"nlocus"} - 1) * 2) .
          ": number of recombination values (" .  @{$datainfo{"recomb"}} .
          ") does not correspond with the total loci (" .
          $datainfo{"nlocus"} . ")\n");
    }
    foreach(@{$datainfo{"recomb"}}) {
      if($_ == 0) {
        die("$datafile,$lineno: a locus distance cannot be equal to 0\n");
      }
    }
  } else {
    die("$datafile,$lineno: recombination values between the markers must " .
        "be listed\n");
  }

  close(DATAFILE)
      or warn "Couldn't close $datafile: $!\n";
}

sub parsePedfile {
  my($i, $line) = 0;

  open(PEDFILE, "<$pedfile")
      or die "Couldn't open $pedfile: $!\n";

  $line = 0;
  foreach(<PEDFILE>) {
    $line++;
    if($_ =~ m/^\s*((\d+\s+)+\d+).?\s*Ped:\s*(\S+)\s*Per:\s*(\d+).?$/) {
     $pedinfo[$i++] = [split(/\s+/,$1), $3, $4];
    } else {
      die "Syntax error in $pedfile on line $line\n";
    }
  }

  close(PEDFILE)
      or warn "Couldn't close $pedfile: $!\n";
}

1;
