#!/usr/bin/perl -w
#
#    Copyright 2001,2004 Greenwood Genetic Center
#
#    This file is part of Linkage Scripts.
#
#    Linkage Scripts is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Linkage Scripts is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Linkage Scripts; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use Linkage;

parseDatafile();
parsePedfile();

my(@tmpfiles) = ("datafile.dat",
                 "ipedfile.dat",
                 "pedfile.dat",
                 "speedfile.dat",
                 "tempdat.dat",
                 "tempped.dat");

my(@outfiles) = ("lsp.log",
                 "lsp.stm",
                 "lsp.tmp",
                 "final.dat",
                 "stream.dat",
                 "outfile.dat",
                 "recfile.dat",
                 "final.out",
                 "stream.out");

my($disease) = 1;
my($locus_start) = 2;
my($sendto) = "";
my($big_neg_val) = -1000000;
my(@families) = ();
my($locus, $theta_max, $lod_max, $theta, $title, $pedentry, $newfamily,
   $tempfile);

# Determine the title from the families in the pedfile
foreach $pedentry (@pedinfo) {
  $newfamily = 1; #true
  foreach(@families) {
    if($pedentry->[@$pedentry - 2] eq $_) {
      $newfamily = 0; #false
    }
  }
  if($newfamily == 1) {
    push(@families, $$pedentry[@$pedentry - 2]);
  }
}

if(@families == 1) {
  $title = "Family: $families[0]";
} else {
  $title = "Families: " . join(", ", @families,);
}

# see if we should run in debug mode (e.g. print the output of utilities to
# the screen)
if(exists($ARGV[0]) && $ARGV[0] eq "DEBUG") {
  $sendto = "";
} else {
  $tempfile = "/tmp/linkage" . int(rand(10 ** 10));
  $sendto = "&> $tempfile";
}

# delete the known temporary files, the previous output files, and the
# temporary wildcard files
foreach((@outfiles,
         @tmpfiles,
         <*.tpd>,
         <*.tdt>,
         <*.mpd>,
         <*.mdt>,
	 <checkpoint*>,
	 <script*>,
	 <outf*>,
	 <main*>)) {

  if(-f $_) {
    unlink($_)
        or die "Couldn't delete $_: $!\n";
  }
}

# Run mlink on each locus
for($locus = $locus_start; $locus <= $datainfo{"nlocus"}; $locus++) {
  printf("Processing marker %-8s (%2d of %2d)\n",
         $datainfo{"names"}->[$locus - 1], $locus, $datainfo{"nlocus"});
  runMlink($disease, $locus);
}

# Process the results of mlink
open(DATAOUT, ">mlink.txt")
    or die "Couldn't open results.txt: $!\n";

open(CSVOUT, ">mlink.csv")
    or die "Couldn't open results.txt: $!\n";

# print the header information
print DATAOUT "$title\n\n";

print CSVOUT "$title\n\n";

print DATAOUT "            Theta  Theta  Theta  Theta  Theta  Theta   " .
              "Theta  Max\n";
print DATAOUT "   Locus    0.001  0.01   0.05   0.10   0.20   0.30    Max    " .
              "LOD Score\n";
print DATAOUT "--------------------------------------------------------------" .
              "----\n";
print CSVOUT "Locus,Theta 0.001,Theta 0.01,Theta 0.05,Theta 0.10,Theta 0.20," .
             "Theta 0.30,Theta Max,Max LOD Score\n";

open(PEDIN, "<final.out")
    or die "Couldn't open final.out: $!\n";

# process each line of the output of mlink (final.out)
foreach(<PEDIN>) {
  if($_ =~ m/^THETAS\s+(\d+\.\d+)$/) { #if this line changes the value of theta
    $theta = $1;
  }
  if($_ =~ m/LOD SCORE =\s+(-?\d+\.\d+)$/) { # if the line contains a lod score 
    if($theta == .001 || $theta == .01 || $theta == .05 || $theta == .1
        || $theta == .2 || $theta == .3) {  #if the line contains a lod score of
      printf DATAOUT "%6.2f ", $1;          #intrest, print it
      printf CSVOUT "%.2f,", $1;
    }
    if($1 > $lod_max) { #if the lod score is bigger than the previously
      $lod_max = $1;    #maximum recorded lod score, store it
      $theta_max = $theta;
    }
  }
  if($_ =~ m/^\s+Locus Order\s+:\s+(\d+)\s+(\d+)$/) { #if we reach a new marker
    $locus = $2;  #set our current locus
    if($locus != $locus_start) {     #if this is not the first locus, output the
      if($lod_max == $big_neg_val) { #theta max and lod max for the last locus
        print DATAOUT " Unknown Locus\n";
      } else {
        printf DATAOUT "%7.2f%7.2f\n", $theta_max, $lod_max;
        printf CSVOUT "%.2f,%.2f\n", $theta_max, $lod_max;
      }
    }
    $lod_max = $big_neg_val; # reset lod max
    printf DATAOUT "%8s  ", $datainfo{"names"}->[$locus - 1]; #print the names
    printf CSVOUT "%s,", $datainfo{"names"}->[$locus - 1];    #of the new locus
  }
}

# print the theta max and lod max for the last marker
if($lod_max == $big_neg_val) {
  print DATAOUT " Unknown Locus\n";
  print CSVOUT "Unknown Locus\n";
} else {
  printf DATAOUT "%7.2f%7.2f\n", $theta_max, $lod_max;
  printf CSVOUT "%.2f,%.2f\n", $theta_max, $lod_max;
}

# give the output a timestamp
print DATAOUT "\n\nProcessed on " . localtime(time()) . "\n";
print CSVOUT  "\n\nProcessed on " . localtime(time()) . "\n";

# close the file handles
close(PEDIN)
    or warn "Couldn't close pedin.dat: $!\n";
close(DATAOUT)
    or warn "Couldn't close mlink.txt: $!\n";
close(CSVOUT)
    or warn "Couldn't close mlink.csv: $!\n";

# clean up our tempfiles and the temporary wildcard files
foreach((@tmpfiles,
         <*.tpd>,
         <*.tdt>,
         <*.mpd>,
         <*.mdt>,
	 <checkpoint*>,
	 <script*>,
	 <outf*>,
	 <main*>)) {
  if(-f $_) {
    unlink($_)
        or die "Couldn't delete $_: $!\n";
  }
}

# output the results
print "Results:\n";
system("cat mlink.txt");

sub runMlink {
  system("cp pedin.dat pedin.tpd");
  system("cp datain.dat datain.tdt");

  if(system("lsp " .
            "mlink " .
            "pedin.tpd datain.tdt " .
            "2 " .
            "$_[0]  $_[1] " .
            "0 " .
            "0 " .
            "1 0.001 0.5 0") != 0) {

    die "\nmlink.pl:  error running lsp; see output above\n";
  }

  system("cat lsp.log >> final.out");
  system("cat lsp.stm >> stream.out");

  if(system("unknown" . $sendto) != 0) {
    dump_prog_out();
    die "\nmlink.pl:  error running unknown; see output above\n";
  }

  if(system("mlink" . $sendto) != 0) {
    dump_prog_out();
    die "\nmlink.pl:  error running mlink; see output above\n";
  }

  system("cat outfile.dat >> final.out");
  system("cat stream.dat  >> stream.out");
}


##########################################################################
##  dump_prog_out()
##    dumps the error message of the last program run
##########################################################################
sub dump_prog_out {
    if($tempfile) {
	open(LINKOUT, "$tempfile") or die "Couldn't open $tempfile: $!\n";

	foreach(<LINKOUT>) {
	    print $_;
	}

	close(LINKOUT) or warn "Unable to close $tempfile: $!\n";
    }
}

END {
    if($tempfile) {
	unlink("$tempfile") or warn "unable to unlink $tempfile: $!\n";
    }
}
